local Brainfuck
do
  local _class_0
  local _base_0 = {
    maximum_runtime = 16384,
    shift_right = function(self)
      self.memory_pointer = self.memory_pointer + 1
      if self.matrix[self.memory_pointer] == nil then
        self.matrix[self.memory_pointer] = 0
      end
    end,
    shift_left = function(self)
      self.memory_pointer = self.memory_pointer - 1
      if self.matrix[self.memory_pointer] == nil then
        self.matrix[self.memory_pointer] = 0
      end
    end,
    increment = function(self)
      self.matrix[self.memory_pointer] = (self.matrix[self.memory_pointer] + 1) % 256
    end,
    decrement = function(self)
      self.matrix[self.memory_pointer] = (self.matrix[self.memory_pointer] - 1) % 256
    end,
    jez_forward = function(self)
      if self:get_value() == 0 then
        local depth = 0
        while true do
          self:move_forward()
          local _exp_0 = self:get_current_char()
          if "[" == _exp_0 then
            depth = depth + 1
          elseif "]" == _exp_0 then
            if depth == 0 then
              return nil
            else
              depth = depth - 1
            end
          elseif nil == _exp_0 then
            return nil
          end
        end
      end
    end,
    jnz_backward = function(self)
      if self:get_value() ~= 0 then
        local depth = 0
        while true do
          self:move_backward()
          local _exp_0 = self:get_current_char()
          if "[" == _exp_0 then
            if depth == 0 then
              return nil
            else
              depth = depth - 1
            end
          elseif "]" == _exp_0 then
            depth = depth + 1
          elseif "" == _exp_0 then
            return nil
          end
        end
      end
    end,
    read_input = function(self)
      local char = string.sub(self.input, 1, 1)
      if char ~= "" then
        self:set_value(string.byte(char))
      else
        self:set_value(0)
      end
      self.input = string.sub(self.input, 2)
    end,
    move_forward = function(self)
      self.code_pointer = self.code_pointer + 1
    end,
    move_backward = function(self)
      self.code_pointer = self.code_pointer - 1
    end,
    get_char = function(self, n)
      return string.sub(self.code, n, n)
    end,
    get_current_char = function(self)
      return self:get_char(self.code_pointer)
    end,
    get_value = function(self)
      return self.matrix[self.memory_pointer]
    end,
    set_value = function(self, value)
      self.matrix[self.memory_pointer] = value
    end,
    append_input = function(self, input)
      self.input = self.input .. input
    end,
    step = function(self)
      local ret = nil
      local _exp_0 = self:get_current_char()
      if "[" == _exp_0 then
        self:jez_forward()
      elseif "]" == _exp_0 then
        self:jnz_backward()
      elseif "+" == _exp_0 then
        self:increment()
      elseif "-" == _exp_0 then
        self:decrement()
      elseif "." == _exp_0 then
        ret = string.char(self:get_value())
        if self.output_callback ~= nil then
          self:output_callback(ret)
        end
      elseif "," == _exp_0 then
        self:read_input()
      elseif ">" == _exp_0 then
        self:shift_right()
      elseif "<" == _exp_0 then
        self:shift_left()
      elseif "" == _exp_0 then
        return false
      end
      self:move_forward()
      return ret
    end,
    run = function(self)
      local output = ""
      for i = 1, self.maximum_runtime do
        local step = self:step()
        if step == false then
          break
        elseif step ~= nil then
          output = output .. step
        end
      end
      return output
    end
  }
  _base_0.__index = _base_0
  _class_0 = setmetatable({
    __init = function(self, code, input, output_callback)
      self.code = code
      self.input = input
      self.output_callback = output_callback
      self.matrix = {
        0
      }
      self.memory_pointer = 1
      self.code_pointer = 1
    end,
    __base = _base_0,
    __name = "Brainfuck"
  }, {
    __index = _base_0,
    __call = function(cls, ...)
      local _self_0 = setmetatable({}, _base_0)
      cls.__init(_self_0, ...)
      return _self_0
    end
  })
  _base_0.__class = _class_0
  Brainfuck = _class_0
end
_BF = { }
_BF.programs = { }
addhook("always", "_BF.step_programs")
_BF.step_programs = function()
  for key, program in pairs(_BF.programs) do
    _BF.current_program = program
    local step = program.instance:step()
    if step == false then
      table.remove(_BF.programs, key)
    elseif step == "\0" then
      local write_input = (string.sub(program.output_buffer, 1, 1)) == "!"
      program.output_buffer = write_input and (string.sub(program.output_buffer, 2)) or program.output_buffer
      local ret = (loadstring(program.output_buffer))()
      if write_input then
        local _exp_0 = type(ret)
        if "string" == _exp_0 then
          program.instance:append_input(ret)
        elseif "number" == _exp_0 then
          program.instance:append_input(string.char(ret))
        elseif "boolean" == _exp_0 then
          program.instance:append_input((ret and "\1" or "\0"))
        else
          program.instance:append_input(tostring(ret))
        end
      end
      program.output_buffer = ""
    elseif step ~= nil then
      program.output_buffer = program.output_buffer .. step
    end
  end
end
_BF.add_program = function(program, input, callback)
  input = (input or "")
  local program_table = {
    output_buffer = "",
    instance = Brainfuck(program, input, callback)
  }
  return table.insert(_BF.programs, program_table)
end
_BF.dofile = function(file, input, callback)
  local f = io.open(file, "rb")
  local content = f:read("*all")
  f:close()
  return _BF.add_program(content, input, callback)
end
