local TISfuck
do
  local _class_0
  local _base_0 = {
    maximum_runtime = 16384,
    shift_right = function(self)
      local num = self:try_read_number()
      self.memory_pointer = self.memory_pointer + (num or 1)
      if self.matrix[self.memory_pointer] == nil then
        self.matrix[self.memory_pointer] = 0
      end
    end,
    shift_left = function(self)
      local num = self:try_read_number()
      self.memory_pointer = self.memory_pointer - (num or 1)
      if self.matrix[self.memory_pointer] == nil then
        self.matrix[self.memory_pointer] = 0
      end
    end,
    increment = function(self)
      local num = self:try_read_number()
      return self:set_value((self:get_value() + (num or 1)) % 256)
    end,
    decrement = function(self)
      local num = self:try_read_number()
      return self:set_value((self:get_value() - (num or 1)) % 256)
    end,
    jump_forward = function(self)
      local num = self:try_read_number(true)
      if self:get_value() == (num or 0) then
        local depth = 0
        while true do
          if self:move_forward() then
            return nil
          end
          local _exp_0 = self:get_current_char()
          if "[" == _exp_0 then
            depth = depth + 1
          elseif "]" == _exp_0 then
            if depth == 0 then
              self:try_read_number()
              return nil
            else
              depth = depth - 1
            end
          elseif nil == _exp_0 then
            return nil
          end
        end
      else
        return self.try_read_number
      end
    end,
    jump_backward = function(self)
      local num = self:try_read_number(true)
      if self:get_value() ~= (num or 0) then
        local depth = 0
        while true do
          if self:move_backward() then
            return nil
          end
          local _exp_0 = self:get_current_char()
          if "[" == _exp_0 then
            if depth == 0 then
              self:try_read_number()
              return nil
            else
              depth = depth - 1
            end
          elseif "]" == _exp_0 then
            depth = depth + 1
          elseif nil == _exp_0 then
            return nil
          end
        end
      else
        return self.try_read_number
      end
    end,
    halt = function(self)
      local num = self:try_read_number()
      if (num ~= nil) then
        return self:get_value() == num
      end
      return true
    end,
    set = function(self)
      local num = self:try_read_number()
      return self:set_value((num or 0))
    end,
    set_tranferable = function(self)
      local num = self:try_read_number()
      self.tranferable_memory = (num == nil and self:get_value() or num) % 256
    end,
    jump_upward = function(self)
      local num = self:try_read_number()
      return self:set_line(self.line_pointer - (num or 1))
    end,
    jump_downward = function(self)
      local num = self:try_read_number()
      return self:set_line(self.line_pointer + (num or 1))
    end,
    swap_memory = function(self)
      local buf = self:get_value()
      self:set_value(self.tranferable_memory)
      self.tranferable_memory = buf
    end,
    try_read_number = function(self, move_back)
      local line, col = self.line_pointer, self.col_pointer
      if self:move_forward() ~= true then
        local cur_char = self:get_current_char()
        if cur_char == "~" then
          if (move_back) then
            self.line_pointer, self.col_pointer = line, col
          else
            self:move_forward()
          end
          return get_value
        elseif cur_char == "?" then
          if (move_back) then
            self.line_pointer, self.col_pointer = line, col
          else
            self:move_forward()
          end
          return self.tranferable_memory
        elseif (string.find(cur_char, "[A-F0-9~?]")) ~= nil then
          local num = cur_char
          if self:move_forward() ~= true then
            if (string.find(self:get_current_char(), "[A-F,0-9]")) ~= nil then
              num = num .. self:get_current_char()
              self:move_forward()
            end
          end
          if (move_back) then
            self.line_pointer, self.col_pointer = line, col
          end
          return tonumber(num, 16)
        else
          if (move_back) then
            self.line_pointer, self.col_pointer = line, col
          end
          return nil
        end
      else
        if (move_back) then
          self.line_pointer, self.col_pointer = line, col
        end
        return nil
      end
    end,
    read_input = function(self)
      local char = string.sub(self.input, 1, 1)
      if char ~= "" then
        self:set_value(string.byte(char))
      else
        self:set_value(0)
      end
      self.input = string.sub(self.input, 2)
    end,
    move_forward = function(self)
      self.col_pointer = self.col_pointer + 1
      if self:get_current_char() == nil then
        self:set_line(self.line_pointer + 1)
        return true
      end
    end,
    move_backward = function(self)
      self.col_pointer = self.col_pointer - 1
      if self:get_current_char() == nil then
        self:set_line(self.line_pointer - 1)
        return true
      end
    end,
    set_line = function(self, l)
      self.line_pointer = ((l - 1) % #self.code) + 1
      self.col_pointer = 1
    end,
    get_current_line = function(self, l)
      return self:get_line(self.line_pointer)
    end,
    get_line = function(self, l)
      return self.code[((l - 1) % #self.code) + 1]
    end,
    get_char = function(self, l, c)
      if (c > #self.code[l] or c < 1) then
        return nil
      else
        return string.sub(self.code[l], c, c)
      end
    end,
    get_current_char = function(self)
      return self:get_char(self.line_pointer, self.col_pointer)
    end,
    get_value = function(self)
      return self.matrix[self.memory_pointer]
    end,
    set_value = function(self, value)
      self.matrix[self.memory_pointer] = value
    end,
    append_input = function(self, input)
      self.input = self.input .. input
    end,
    step = function(self)
      local ret = nil
      local _exp_0 = self:get_current_char()
      if "[" == _exp_0 then
        self:jump_forward()
      elseif "]" == _exp_0 then
        self:jump_backward()
      elseif "+" == _exp_0 then
        self:increment()
      elseif "-" == _exp_0 then
        self:decrement()
      elseif "." == _exp_0 then
        local num = self:try_read_number()
        ret = string.char((num or self:get_value()))
        if self.output_callback ~= nil then
          self:output_callback(ret)
        end
      elseif "," == _exp_0 then
        self:read_input()
        self:move_forward()
      elseif ">" == _exp_0 then
        self:shift_right()
      elseif "<" == _exp_0 then
        self:shift_left()
      elseif ";" == _exp_0 then
        if self:halt() then
          return false
        end
      elseif "*" == _exp_0 then
        self:set()
      elseif "!" == _exp_0 then
        self:set_tranferable()
      elseif "^" == _exp_0 then
        self:jump_upward()
      elseif "v" == _exp_0 then
        self:jump_downward()
      elseif ":" == _exp_0 then
        self:swap_memory()
        self:move_forward()
      elseif nil == _exp_0 then
        return false
      end
      return ret
    end,
    run = function(self)
      local output = ""
      for i = 1, self.maximum_runtime do
        local step = self:step()
        if step == false then
          break
        elseif step ~= nil then
          output = output .. step
        end
      end
      return output
    end
  }
  _base_0.__index = _base_0
  _class_0 = setmetatable({
    __init = function(self, code, input, output_callback)
      self.code = code
      self.input = input
      self.output_callback = output_callback
      self.matrix = {
        0
      }
      self.memory_pointer = 1
      self.col_pointer = 1
      self.line_pointer = 1
      self.tranferable_memory = 0
    end,
    __base = _base_0,
    __name = "TISfuck"
  }, {
    __index = _base_0,
    __call = function(cls, ...)
      local _self_0 = setmetatable({}, _base_0)
      cls.__init(_self_0, ...)
      return _self_0
    end
  })
  _base_0.__class = _class_0
  TISfuck = _class_0
end
local split_string
split_string = function(input, sep)
  sep = sep or '%s'
  local t = { }
  for field, s in (string.gmatch(input, "([^" .. sep .. "]*)(" .. sep .. "?)")) do
    table.insert(t, field)
    if s == "" then
      return t
    end
  end
end
_TF = { }
_TF.programs = { }
addhook("always", "_TF.step_programs")
_TF.step_programs = function()
  for key, program in pairs(_TF.programs) do
    _TF.current_program = program
    local step = program.instance:step()
    if step == false then
      table.remove(_TF.programs, key)
    elseif step == "\0" then
      local write_input = (string.sub(program.output_buffer, 1, 1)) == "!"
      program.output_buffer = write_input and (string.sub(program.output_buffer, 2)) or program.output_buffer
      local ret = (loadstring(program.output_buffer))()
      if write_input then
        local _exp_0 = type(ret)
        if "string" == _exp_0 then
          program.instance:append_input(ret)
        elseif "number" == _exp_0 then
          program.instance:append_input(string.char(ret))
        elseif "boolean" == _exp_0 then
          program.instance:append_input((ret and "\1" or "\0"))
        else
          program.instance:append_input(tostring(ret))
        end
      end
      program.output_buffer = ""
    elseif step ~= nil then
      program.output_buffer = program.output_buffer .. step
    end
  end
end
_TF.add_program = function(program, input, callback)
  input = (input or "")
  local program_table = {
    output_buffer = "",
    instance = TISfuck(program, input, callback)
  }
  return table.insert(_TF.programs, program_table)
end
_TF.dofile = function(file, input, callback)
  local f = io.open(file, "rb")
  local content = f:read("*all")
  f:close()
  local program = split_string((string.gsub(content, "\r", "")), "\n")
  return _TF.add_program(program, input, callback)
end
